# Copyright (c) 2004-2017  Pavel Rychly, Vojtech Kovar, Milos Jakubicek, Vit Baisa

import CGIPublisher
import os, json
from jobrunner import JobClient
 
def load_opt_file (options, filepath, option_list=[], selector=''):
    if not os.path.isfile (filepath):
        return
    read_options = {}
    for line in open (filepath).readlines():
        if line[0] == '#':
            continue
        try:
            a, v = line.split ('\t', 1)
        except ValueError:
            continue
        attr = a.strip()
        read_options[attr] = v.strip('\n').decode('utf8')
        if selector: CGIPublisher.choose_selector(read_options, selector)
        for k, v in read_options.items():
            if not option_list or k in option_list:
                if v.startswith(("{", "[")): options[k] = json.loads(v)
                else: options[k] = v

def set_inner_value(options, path, new_value):
    # throws IndexError, KeyError, TypeError
    path_list = path.split('.')
    path0_list = path_list[0].split('[')
    option = path0_list[0]
    whole_value = options.get(option, {})
    curr_value = whole_value
    curr_key = None
    for it in path0_list[1:] + path_list[1:]: # find the object to set
        for item in it.split('['):
            if curr_key is None: next_value = curr_value
            else: next_value = curr_value[curr_key]
            if item == '__delete':
                if curr_key is None: whole_value = ''
                else: del curr_value[curr_key]
            elif item == '__append':
                if curr_key is None:
                    if whole_value == {}: whole_value = []
                    whole_value.append(new_value)
                else:
                    if curr_value[curr_key] == {}: curr_value[curr_key] = []
                    curr_value[curr_key].append(new_value)
            elif item.endswith(']'): # number
                num = int(item[:-1])
                if not next_value:
                    if curr_key is None:
                        next_value = whole_value = [{}]
                    else:
                        curr_value[curr_key] = [{}]
                curr_value = next_value
                curr_key = num
            else: # key
                if not item in next_value:
                    next_value[item] = {}
                curr_value = next_value
                curr_key = item
    if type(0)==type(curr_key) and type({})==type(curr_value):
        raise IndexError('int in dict not allowed here')
    if item not in ('__delete', '__append'):
        curr_value[curr_key] = new_value
    return option, whole_value


class UserCGI (CGIPublisher.CGIPublisher):
    _ca_user_info = u''
    _ca_api_info = u''
    _ca_languages = []
    _options_dir = u''
    _email = u''
    _default_user = u'defaults'
    _superuser = False
    _admin_email = ''
    _job_server = 'localhost'
    _job_port = 8333
    _job_dir = u''
    _data_dir = u''
    # list of usernames granted access to job management
    _superusers = []
    attrs2save = []
    display_adsense = False

    def __init__ (self, user=None):
        CGIPublisher.CGIPublisher.__init__ (self)
        self._user = user
        if not self._admin_email:
            self._admin_email = os.environ.get("SERVER_ADMIN", "root@localhost")
        if not self._job_dir:
            self._job_dir = os.path.dirname (os.path.abspath(self._cache_dir))\
                                             + "/jobs"
        if not os.path.isdir(self._job_dir):
            os.makedirs(self._job_dir)
        self._job = JobClient (self._job_server, self._job_port, self._job_dir,\
                               self._admin_email, self._email, self._user,
                               self._superuser)

    def _user_defaults (self, user):
        pass

    def _get_ca_corpnames(self, language='', with_ws=0):
        import urllib2, simplejson
        request = urllib2.Request(self._ca_corp_info
                % (self._sessionid, self._user_ip, language, with_ws))
        file = urllib2.urlopen(request)
        data = file.read()
        return simplejson.loads(data)

    def _get_ca_languages(self):
        import urllib2, simplejson
        request = urllib2.Request(self._ca_lang_info)
        file = urllib2.urlopen(request)
        data = file.read()
        self._ca_languages = simplejson.loads(data)

    api_key = ''
    url_username = ''

    def _get_ca_auth_request(self, corpname):
        import urllib2
        if self._ca_user_info and corpname:
            return urllib2.Request(self._ca_user_info
                                  % (self._sessionid, self._user_ip, corpname))
        elif self._ca_api_info:
            usr = self.url_username
            passwd = self.api_key
            return urllib2.Request(self._ca_api_info % (usr, passwd, corpname))
        else:
            return urllib2.Request(self._ca_user_access
                                   % (self._sessionid, self._user_ip))

    def _get_ca_user_info(self, corpname=''):
        import urllib, urllib2, simplejson
        if self._login_address:
            self._login_address += '/?next=' + urllib.quote_plus(
                                                os.getenv ('REQUEST_URI', '-'))
        request = self._get_ca_auth_request(corpname)
        no_tries = 0
        while True:
            try:
                file = urllib2.urlopen(request)
                break
            except urllib2.HTTPError:
                import time
                if no_tries <= 5: time.sleep(3); no_tries += 1
                else: raise
        data = file.read()
        file.close()
        user_info = simplejson.loads(data)
        if user_info.get('_authenticated', False): self._has_access = True
        if user_info.has_key('_username'): self._user = user_info['_username']
        if user_info.has_key('error'): self._has_access = False
            # set attributes (incl. these starting '_') and correct types
        corr_func = {type(0): int, type(0.0): float,
                     type([]): lambda x: x if type(x) == type([]) else [x]}
        for k, v in user_info.items():
            try: setattr (self, k, corr_func[type(getattr(self, k))](v))
            except: setattr (self, k, v)
        if self._user: self._job.user = self._user = str(self._user)
        if self._email.startswith('sso_'): self._email = '' # SSO has fake email
        if self._email: self._job.user_email = self._email
        self._job.superuser = self._superuser

    def _setup_user (self, corpname=''):
        if self._ca_user_info or self._ca_api_info:
            self._has_access = 0
            self._get_ca_user_info(corpname)
        if hasattr(self, '_ca_lang_info'):
            self._get_ca_languages()
        if not self._user:
            self._job.user = self._user = os.getenv ('REMOTE_USER','-')
            if self._user == '-':
                self._user = self._default_user
        if self._user in self._superusers: # not needed when CA
            self._job.superuser = self._superuser = True
        user = self._user
        options = {}
        load_opt_file (options, os.path.join (self._options_dir,
                                              self._default_user))
        if user is not self._default_user:
            load_opt_file (options, os.path.join (self._options_dir, user))
        CGIPublisher.correct_types (options, self.defaults, selector=1)
        self._user_defaults (user)
        self.__dict__.update (options)

    def reset_user_options(self, corpus=''):
        opt_filepath = os.path.join(self._options_dir, self._user)
        if not os.path.exists(opt_filepath):
            return {'status': 'OK'}
        if not corpus: # delete everything
            open(opt_filepath, 'w').close()
            return {'status': 'OK'}
        filt = corpus + ':'
        towrite = []
        source = open(opt_filepath)
        target = open(opt_filepath + '.tmp.copy', 'w')
        for line in source:
            if not line.decode('utf8').startswith(filt):
                target.write(line)
        source.close()
        target.close()
        os.rename(opt_filepath + '.tmp.copy', opt_filepath)
        return {'status': 'OK'}

    def get_user_options(self, options=[], corpus=''):
        self.format = 'json'
        defaults = dict([(opt, self.defaults[opt]) for opt in options
                                                   if opt in self.defaults])
        user_opts = {}
        opt_filepath = os.path.join (self._options_dir, self._user)
        load_opt_file(user_opts, opt_filepath, options, corpus)
        return {'default': defaults, 'user': user_opts}

    def set_user_options(self, options={}, corpus=''):
        import re
        def natural_key(x):
            return [ int(s) if s.isdigit() else s
                     for s in re.split(r'(\d+)', x[0]) ]
        self.format = 'json'
        opt_filepath = os.path.join (self._options_dir, self._user)
        result_options = {}
        load_opt_file(result_options, opt_filepath)
        for k, v in sorted(options.items(), key=natural_key, reverse=True):
            if '.' in k or '[' in k: # parse json path
                try:
                    k, v = set_inner_value(result_options, k, v)
                except (IndexError, KeyError, TypeError):
                    return {'error': k}
            if corpus: key = corpus + ':' + k
            else: key = k
            if (key in result_options and v == self.defaults.get(k, '')):
                del result_options[key]
            elif v != self.defaults.get(k, ''): # save only non-default
                result_options[key] = v
        if not os.path.isdir(self._options_dir):
            os.makedirs(self._options_dir)
        target = open(opt_filepath + '.tmp.copy', 'w')
        for k, v in sorted(result_options.items()):
            if type({}) == type(v) or type([]) == type(v):
                v = json.dumps(v)
            if isinstance(v, unicode): v = v.encode('utf8')
            target.write(str(k) + '\t' + str(v) + '\n')
        target.close()
        os.rename(opt_filepath + '.tmp.copy', opt_filepath)
        return {'status': 'OK'}

    def _save_options (self, optlist=[], selector=''):
        # obsolete, Platypus will not use this
        if selector:
            tosave = [(selector + ':' + opt, self.__dict__[opt])
                           for opt in optlist if self.__dict__.has_key (opt)]
        else:
            tosave = [(opt, self.__dict__[opt]) for opt in optlist
                           if self.__dict__.has_key (opt)]
        opt_filepath = os.path.join (self._options_dir, self._user)
        options = {}
        load_opt_file(options, opt_filepath)
        options.update(tosave)

        opt_lines = []
        for k, v in options.iteritems():
            if type({}) == type(v) or type([]) == type(v):
                v = json.dumps(v)
            if isinstance(v, unicode): v = v.encode('utf8')
            opt_lines.append(str(k) + '\t' + str(v))
        opt_lines.sort()
        if not os.path.isdir(self._options_dir):
            os.makedirs(self._options_dir)
        opt_file = open(opt_filepath, 'w')
        opt_file.write('\n'.join(opt_lines))
        opt_file.close()

    def save_global_attrs(self):
        options = [a for a in self.attrs2save if not a.startswith('_')]
        self._save_options(options, '')
        return _('Selected options successfully saved')

    def send_mail(self, subject, msg, sender='', recipient='', attachment='', sender_name='', cc=''):
        import smtplib
        import urllib
        if not recipient:
            recipient = self._email
        if not sender:
            sender = self._admin_email
        server = smtplib.SMTP('localhost')
        recipients = [recipient, cc]
        msg = urllib.unquote(msg).encode('utf-8')
        from email.MIMEMultipart import MIMEMultipart
        from email.MIMEBase import MIMEBase
        from email.MIMEText import MIMEText
        from email.Utils import formatdate
        from email import Encoders
        m = MIMEMultipart('alternative')
        m['From'] = 'Sketch Engine Feedback <%s>' % sender
        m['To'] = recipient
        m['Sender'] = sender
        if cc:
            m['Cc'] = cc
            m['Reply-To'] = cc
        m['Date'] = formatdate(localtime=True)
        m['Subject'] = subject
        m['Content-Type'] = 'text/html; charset=utf-8'
        m.attach(MIMEText(msg, _charset="utf-8"))
        if attachment[0]:
            part = MIMEBase('application', "octet-stream")
            part.set_payload(attachment[1].read())
            Encoders.encode_base64(part)
            part.add_header('Content-Disposition', 'attachment; filename="%s"'%\
                    attachment[0])
            m.attach(part)
        server.sendmail(sender, recipients, m.as_string())
        server.quit()

    def feedback(self, feedback_url='', navigator='', feedback_corpname='',
            feedback_text='', attachment=('', None), feedback_fullname='',
            feedback_error='', feedback_email=''):
        if not navigator:
            return ""
        import time
        propemail = self._email or feedback_email or 'anonymous@sketchengine.co.uk'
        propname = getattr(self, '_ca_full_name', self._user)
        msg = """%s

USER: %s (%s)
EMAIL: %s
CORPUS: %s
ERR: %s
URL: %s
BROWSER: %s

PROFILE: %s/admin/ca/userprofile/_%s/
LOGIN AS: %s/auth/login_as/?username=%s
""" % \
                (feedback_text, self._user,
                feedback_fullname or propname, propemail,
                feedback_corpname, feedback_error, feedback_url, navigator,
                self._home_url, self._user, self._home_url, self._user)
        self.send_mail('Sketch Engine feedback from %s, %s' %\
                (propname, time.strftime('%Y-%m-%d %H:%M:%S')), msg,
                'webnobody@sketchengine.co.uk',
                self._admin_email, attachment, propname, propemail)
        return _('Your feedback has been sent to %s and in a copy to you. Thank you for your time!'\
                 % self._admin_email)

    def jobs (self, finished=False):
        jobs = self._job.request ("list_user_jobs",
                                  {"user" : self._user, "finished" : finished})
        jobs = json.loads(jobs[0])
        out = {'jobs': jobs, 'view_type' : 'user', 'no_corpus_show': 1}
        return out

    def all_jobs (self, finished=False):
        if not self._superuser:
            raise Exception('access denied: User "%s" is not a superuser' \
                             % self._user)
        jobs = self._job.request ("list_all_jobs",
                                  {"user" : self._user, "finished" : finished})
        jobs = json.loads(jobs[0])
        out = {'jobs': jobs, 'view_type' : 'all', 'no_corpus_show': 1}
        return out
    all_jobs.template = "jobs.tmpl"

    def jobproxy (self, task=""):
        qs = self.__dict__["environ"]["QUERY_STRING"]
        disallowed_tasks = ["new_job", "list_all_jobs", "list_user_jobs"]
        if task in disallowed_tasks:
            raise Exception('access denied: task not allowed for proxy')
        import re, json
        qs = re.sub ("task=[^;&]+[&;]?", "", qs)
        self.format = 'json'
        jr = self._job.request(task, qs)
        return {'signal': jr[0], 'code': jr[1]}

    def get_url (self):
        scheme = self.environ.get ("REQUEST_SCHEME")
        if not scheme:
            scheme = self.environ.get ("HTTPS") and "https" or "http"
        return "%s://%s%s" % (scheme, self.environ["SERVER_NAME"],
                                      self.environ["REQUEST_URI"])
